/*
 * Copyright 2021 Associated Universities, Inc.
 *
 * This file is part of Telescope Time Allocation Tools (TTAT).
 *
 * TTAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * TTAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TTAT.  If not, see <https://www.gnu.org/licenses/>.
 *
 */
import {fakeAsync, TestBed} from '@angular/core/testing';
import {AlertMessage, AlertService} from './alert.service';
import {ReplaySubject} from "rxjs";
import {NavigationEnd, NavigationStart, Router, RouterEvent} from "@angular/router";

describe('AlertService', () => {
  let service: AlertService;
  const eventSubject = new ReplaySubject<RouterEvent>(1);
  const routerMock = {
    navigate: jasmine.createSpy('navigate'),
    events: eventSubject.asObservable(),
    url: 'test/url'
  }

  beforeEach(() => {
    TestBed.configureTestingModule({
      providers: [{"provide": Router, "useValue": routerMock}]
    });
    service = TestBed.inject(AlertService);
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  it('success should change the subject', fakeAsync(() => {
    const sub = service.getMessage().subscribe((message) => {
      expect(message).toEqual({type: 'success', text: 'success message'});
    });
    service.success('success message');
    sub.unsubscribe();
  }));

  it('error should change the subject', fakeAsync(() => {
    const sub = service.getMessage().subscribe((message) => {
      expect(message).toEqual({type: 'error', text: 'error message'});
    });
    service.error('error message');
    sub.unsubscribe();
  }));

  it('info should change the subject', fakeAsync(() => {
    const sub = service.getMessage().subscribe((message) => {
      expect(message).toEqual({type: 'info', text: 'info message'});
    });
    service.info('info message');
    sub.unsubscribe();
  }));

  it('should detect router events and respond based on keep settings', fakeAsync(() => {
    let message = null;
    let counter = 0;
    const expected = {type: 'info', text: 'info message'} as AlertMessage;
    const sub = service.getMessage().subscribe((m: AlertMessage) => {
      counter = counter + 1;
      message = m;
    });
    service.info('info message', true);
    // this event shouldn't trigger anything, so it shouldn't change the message
    eventSubject.next(new NavigationEnd(1,'ignored', 'redirectUrl'));
    // @ts-ignore
    expect(message).toEqual(expected);
    // this event should trigger the change, but as keepAfterNavigationChange is set, the message shouldn't change
    eventSubject.next(new NavigationStart(2,'trigger'));
    service.getMessage().subscribe((m: AlertMessage) => {
      message = m;
    });
    // @ts-ignore
    expect(message).toEqual(expected);

    // this event should trigger the change, but as keepAfterNavigationChange is set, the message shouldn't change
    eventSubject.next(new NavigationStart(3,'trigger'));
    service.getMessage().subscribe((m: AlertMessage) => {
      message = m;
    });
    expect(message).toBeUndefined();
    sub.unsubscribe();
  }));
});
