#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Tests for the execution manager. Only one is feasible; other tests are broken need to be fixed Some Fine Day.
"""

# pylint: disable=E0401, W0212, W0511, W0621

import logging
import sys
from typing import Dict, List
from unittest.mock import patch

import pytest
from _pytest.logging import LogCaptureFixture
from messaging.messenger import SenderIF
from messaging.router import Router

from workspaces.capability.enums import CapabilityRequestState, QueueState
from workspaces.capability.schema import CapabilityExecution, CapabilityRequest
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.workflow.schema import WorkflowRequest
from workspaces.workflow.services.workflow_service import WorkflowService

pytest_plugins = ["testing.utils.conftest"]

logger = logging.getLogger("execution_manager")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def reset_execution_mgr(ex_man: ExecutionManager) -> ExecutionManager:
    """
    Refresh this execution manager.

    :param ex_man: the execution manager
    :return: a brand-spanking-new exec manager
    """
    ex_man.queues = {}
    ex_man.execution_pool = []
    return ex_man


@pytest.fixture(scope="function")
def dummy_execution_manager(
    mock_capability_info: CapabilityInfo,
    mock_workflow_service: WorkflowService,
    test_router: Router,
    test_message_sender: SenderIF,
) -> ExecutionManager:
    """
    Pytest fixture that instantiates a dummy ExecutionManager object for testing purposes

    :return: Dummy ExecutionManager
    """
    return ExecutionManager(mock_capability_info, mock_workflow_service, test_router, test_message_sender)


@pytest.fixture(scope="function")
def workflow_executing_message(
    mock_workflow_service: WorkflowService,
    mock_workflow_requests: List[WorkflowRequest],
) -> Dict:
    """
    Create a reusable workflow-executing message

    :return: Workflow-executing message
    """
    request = mock_workflow_requests[0]
    return {
        "service": "workflow",
        "routing_key": "workflow",
        "type": "workflow-executing",
        "subject": request.__json__(),
    }


@pytest.fixture(scope="function")
def workflow_complete_message(
    mock_workflow_service: WorkflowService, mock_workflow_requests: List[WorkflowRequest]
) -> Dict:
    """
    Create a reusable workflow-complete message

    :return: Workflow-complete message
    """
    request = mock_workflow_requests[0]
    return {
        "service": "workflow",
        "routing_key": "workflow",
        "type": "workflow-complete",
        "subject": request.__json__(),
    }


@pytest.fixture(scope="function")
def step_complete_message(mock_capability_execution: CapabilityExecution) -> Dict:
    """
    Build a capability "step-complete" message

    :param mock_capability_execution:
    :return:
    """
    return {
        "service": "capability",
        "routing_key": "capability",
        "subject": mock_capability_execution.__json__(),
        "type": "step-complete",
    }


# TODO: rewrite test so it works
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_run_capability(
    mock_capability_info: CapabilityInfo,
    dummy_execution_manager: ExecutionManager,
    mock_capability_requests: List[CapabilityRequest],
):
    """
    Tests that run_capability correctly creates an execution given a request
    and enqueues that execution if the first capability step is running a workflow
    """
    null_capability_request = mock_capability_requests[0]
    assert null_capability_request.state == CapabilityRequestState.Created.name
    execution = mock_capability_info.create_execution(null_capability_request)

    # logger.info(f">>> the capability request: {null_capability_request}")
    null_execution = dummy_execution_manager.run_capability(execution)
    assert null_capability_request.state == CapabilityRequestState.Submitted.name
    assert null_execution.state == CapabilityRequestState.Created.name
    assert len(dummy_execution_manager.queues) == 1
    dummy_execution_manager = reset_execution_mgr(dummy_execution_manager)

    # Test with execution that will be placed into the execution pool rather than enqueued
    test_capability_request = mock_capability_requests[1]
    execution = mock_capability_info.create_execution(test_capability_request)

    test_execution = dummy_execution_manager.run_capability(execution)
    assert len(dummy_execution_manager.execution_pool) == 1


# TODO: rewrite test so it works
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_send_step_complete_workflow_executing(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    workflow_executing_message: Dict,
    step_complete_message: Dict,
):
    """
    Tests that sending a workflow-executing step complete message works properly
    """
    with patch("messaging.router.Router.receive_message"):
        dummy_execution_manager.send_step_complete(**workflow_executing_message)
        assert dummy_execution_manager.message_router.receive_message.called_with(step_complete_message)

    # Reset state of mock execution
    mock_capability_execution.current_step_idx = 0


# TODO: rewrite test so it works
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_send_step_complete_workflow_complete(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    workflow_complete_message: Dict,
    step_complete_message: Dict,
):
    """
    Tests that sending a workflow-complete step complete message works properly
    """
    with patch("messaging.router.Router.receive_message"):
        dummy_execution_manager.send_step_complete(**workflow_complete_message)
        assert dummy_execution_manager.message_router.receive_message.called_with(step_complete_message)

    # Reset state of mock execution
    mock_capability_execution.current_step_idx = 0


# TODO: rewrite test so it works
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_on_step_complete(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    step_complete_message: Dict,
):
    """
    Tests that on_step_complete correctly advances and completes executions
    """
    # Simulate starting conditions of queues
    dummy_execution_manager.enqueue_execution(mock_capability_execution)

    with patch("workspaces.capability.services.capability_info.CapabilityInfo.save_execution"):
        # _advance_execution will be called
        assert mock_capability_execution.current_step_idx == 0
        dummy_execution_manager.on_step_complete(**step_complete_message)
        assert mock_capability_execution.current_step_idx == 1
        dummy_execution_manager.capability_info.save_execution.assert_called_with(mock_capability_execution)
        # On AwaitWorkflow step; queues["null"].complete_execution will be called
        # On last step; _complete_execution will be called
        with patch("workspaces.capability.services.capability_queue.CapabilityQueue.complete_execution"):
            with patch("workspaces.capability.services.execution_manager.Router.send_message"):
                assert mock_capability_execution.on_last_step()
                dummy_execution_manager.on_step_complete(**step_complete_message)
                dummy_execution_manager.queues["null"].complete_execution.assert_called_with(
                    mock_capability_execution.id
                )
                dummy_execution_manager.capability_info.save_execution.assert_called_with(mock_capability_execution)
                assert dummy_execution_manager.message_router.send_message.call_count == 1
                assert mock_capability_execution.state == "Complete"

    # Reset state of mock execution
    mock_capability_execution.current_step_idx = 0


@pytest.mark.skip("Broken due to queue/messenger rework")
def test_on_step_complete_unrecognized(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    step_complete_message: Dict,
    caplog: LogCaptureFixture,
):
    """
    Tests that on_step_complete identifies a badly formatted message
    """
    step_complete_message["subject"]["type"] = "Unrecognizable"
    dummy_execution_manager.on_step_complete(**step_complete_message)
    messages = [record.message for record in caplog.records]
    assert "Unrecognized message type: Unrecognizable." in messages


# TODO: rewrite test so it works
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_complete_execution(dummy_execution_manager: ExecutionManager, mock_capability_execution: CapabilityExecution):
    """
    Tests that _complete_execution correctly marks an execution as complete and sends the
    corresponding message
    """
    with patch("messaging.router.Router.receive_message"):
        # Reset execution state
        mock_capability_execution.state = CapabilityRequestState.Created.name

        dummy_execution_manager._complete_execution(mock_capability_execution)
        assert dummy_execution_manager.message_router.receive_message.called_with(
            {
                "service": "capability",
                "routing_key": "capability",
                "subject": mock_capability_execution.__json__(),
                "type": "execution-complete",
            }
        )
        assert mock_capability_execution.state == "Complete"


@pytest.mark.skip(
    reason="Code that utilizes JSON serialization such as MessageArchitect throws Exceptions when called from tests"
)
def test_fail_execution(dummy_execution_manager: ExecutionManager, mock_capability_execution: CapabilityExecution):
    with patch("messaging.router.Router.receive_message"):
        mock_capability_execution.state = CapabilityRequestState.Created.name
        mock_capability_execution.queue_state = QueueState.Queued.name
        updated_execution = dummy_execution_manager._fail_execution(mock_capability_execution)
        assert dummy_execution_manager.router.receive_message.called_with(
            {
                "service": "capability",
                "routing_key": "capability",
                "subject": mock_capability_execution.__json__(),
                "type": "execution-failed",
            }
        )
        assert updated_execution.queue_state == "NotQueued"


# TODO: rewrite test so it works
@pytest.mark.skip("Broken due to queue/messenger rework")
def test_enqueue_execution(dummy_execution_manager: ExecutionManager, mock_capability_execution: CapabilityExecution):
    """
    Tests that executions are properly enqueued
    """
    dummy_execution_manager.execution_pool.append(mock_capability_execution)
    dummy_execution_manager.enqueue_execution(mock_capability_execution)
    assert len(dummy_execution_manager.queues) == 1
    assert len(dummy_execution_manager.queues["null"]) == 1
    assert dummy_execution_manager.queues["null"].queue_manager.start.call_count == 1
    assert mock_capability_execution not in dummy_execution_manager.execution_pool
